/* ----------------------------------------------------------------------------
 * @brief Ethernet Digital Loopback
 * 
 * @copyright Copyright 2019, Octavo Systems, LLC. All rights reserved.
 * 
 * ----------------------------------------------------------------------------
 * 
 * License:     
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its 
 * contributors may be used to endorse or promote products derived from 
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * ----------------------------------------------------------------------------
 * 
 * This program is used to test Ethernet digital loopback on the OSD335x Family 
 * of System-in-Package devices.  See README for instructions on how to use 
 * ths program.
 *
 * This program only uses IEEE defined registers for Ethernet PHYs and 
 * should support all Ethernet PHYs.
 *
 * --------------------------------------------------------------------------*/

// ----------------------------------------------------------------------------
// Includes
// ----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/mman.h>


#include "cpsw.h"
#include "mdio.h"



// ----------------------------------------------------------------------------
// Defines
// ----------------------------------------------------------------------------

#define DEBUG                               0

// Memory defines
#define MAP_SIZE                            4096UL
#define MAP_MASK                           (MAP_SIZE - 1)

// Pinmux registers
#define CONTROL_BASE_ADDR                  (0x44E10000)
#define CONF_XDMA_EVENT_INTR1              (CONTROL_BASE_ADDR + 0x9b4)

// PRCM registers
#define PRCM_BASE_ADDR                     (0x44E00000)
#define CM_CLKOUT_CTRL                     (PRCM_BASE_ADDR + 0x700)

// Ethernet registers
#define SOC_CPSW_SS_REGS                   (0x4A100000)
#define SOC_CPSW_MDIO_REGS                 (0x4A101000)
#define SOC_CPSW_WR_REGS                   (0x4A101200)
#define SOC_CPSW_CPDMA_REGS                (0x4A100800)
#define SOC_CPSW_ALE_REGS                  (0x4A100D00)
#define SOC_CPSW_STAT_REGS                 (0x4A100900)
#define SOC_CPSW_PORT_0_REGS               (0x4A100100)
#define SOC_CPSW_PORT_1_REGS               (0x4A100200)
#define SOC_CPSW_SLIVER_1_REGS             (0x4A100D80)
#define SOC_CPSW_PORT_2_REGS               (0x4A100300)
#define SOC_CPSW_SLIVER_2_REGS             (0x4A100DC0)
#define SOC_CPSW_CPPI_RAM_REGS             (0x4A102000)

// Ethernet Parameter defines
#define ETH_CONTROL_DEFAULT                (0x1100)
#define ETH_CONTROL_LOOPBACK               (0x4000)
#define ETH_CONTROL_SPEED_10               (0x0100)
#define ETH_CONTROL_SPEED_100              (0x2100)
#define ETH_CONTROL_SPEED_1000             (0x0140)


// ----------------------------------------------------------------------------
// Macros
// ----------------------------------------------------------------------------

#define FATAL do { fprintf(stderr, "Error at line %d, file %s (%d) [%s]\n", \
  __LINE__, __FILE__, errno, strerror(errno)); exit(1); } while(0)


      
// ----------------------------------------------------------------------------
// Function Prototypes
// ----------------------------------------------------------------------------

void        * map_address              (int fd, off_t base_addr);
void          unmap_address            (void * virt_base_addr, off_t base_addr);
unsigned long read_write_mem           (int fd, int argc, unsigned int address, char type, unsigned long writeval);

void          print_phy_config         (unsigned int base_addr, unsigned int phy_addr);
void          set_phy_control          (unsigned int base_addr, unsigned int phy_addr, unsigned int loopback, unsigned int speed);
void          print_usage              ();



// ----------------------------------------------------------------------------
// Main
// ----------------------------------------------------------------------------

int main(int argc, char ** argv) {

    int       fd;
    void     *mdio_base_virt;
    void     *cpsw_sliver_1_base_virt;
    void     *cpsw_sliver_2_base_virt;
    
    int       phy_address, phy_loopback, phy_speed;

    // Process command line arguments
    if (argc != 4) {
        print_usage();
        return 0;
    }

    // Convert strings to integers
    phy_address  = atoi(argv[1]);
    phy_loopback = atoi(argv[2]);
    phy_speed    = atoi(argv[3]);
    
    // Check arguments
    if ((phy_address < 0) || (phy_address > 0x1F)) {
        printf("Invalid PHY Address  - Please use a value between 0x0 and 0x1F\n");
        return 0;
    }
    
    if (!((phy_loopback == 1) || (phy_loopback == 0))) {
        printf("Invalid PHY Loopback - Please use a value of 1 for loopback or 0 for normal\n");
        return 0;
    }

    if (!((phy_speed == 100) || (phy_speed == 1000))) {
        printf("Invalid PHY Speed    - Please use a value of 100 or 1000\n");
        return 0;
    }

    if (phy_loopback) {
        printf("Setting PHY at address %d into digital loopback mode with speed %d Mbps\n", phy_address, phy_speed);
    } else {
        printf("Setting PHY at address %d into normal mode (restarts auto-ngotiation)\n", phy_address);
    }


    // Bypass virtual memory
    if ((fd = open("/dev/mem", O_RDWR | O_SYNC)) == -1) FATAL;
    fflush(stdout);
    
    // Map memory pages 
    mdio_base_virt          = map_address(fd, (off_t) SOC_CPSW_MDIO_REGS);    
    cpsw_sliver_1_base_virt = map_address(fd, (off_t) SOC_CPSW_SLIVER_1_REGS);    


    // Print PHY configs
    if (DEBUG) {
        print_phy_config((unsigned int)mdio_base_virt, phy_address);
    }

    // Set both PHYs into loopback mode
    set_phy_control((unsigned int)mdio_base_virt, phy_address, phy_loopback, phy_speed);

    // Set transfer mode to Gigabit
    CPSWSlTransferModeSet((unsigned int)cpsw_sliver_1_base_virt, CPSW_SLIVER_GIG_FULL_DUPLEX);

    // Enable MAC
    CPSWSlRGMIIEnable((unsigned int)cpsw_sliver_1_base_virt);


    // Close memory pages    
    unmap_address(mdio_base_virt,          (off_t)SOC_CPSW_MDIO_REGS);
    unmap_address(cpsw_sliver_1_base_virt, (off_t)SOC_CPSW_SLIVER_1_REGS);

    close(fd);
    
    return 0;
}


void print_usage() {
    printf("Usage:  set_eth_digital_loopback <PHY_ADDRESS> <PHY_LOOPBACK> <PHY_SPEED>\n");
    printf("    PHY_ADDRESS  - MDIO address of Ethernet PHY\n");
    printf("    PHY_LOOPBACK - Loopback mode = 1, Normal mode = 0\n");
    printf("    PHY_SPEED    - Speed of Ethernet PHY in [100, 1000]\n");
}



// ----------------------------------------------------------------------------
// Memory Functions
// ----------------------------------------------------------------------------

void * map_address(int fd, off_t base_addr) {
    void     *map_base; 
    void     *virt_base_addr;
    
    /* Map one page */
    map_base = mmap(0, MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, base_addr & ~MAP_MASK);
    if (map_base == (void *) -1) FATAL;
    if (DEBUG) { printf("Memory mapped at address %p.\n", map_base); }
    fflush(stdout);
    
    virt_base_addr = map_base + (base_addr & MAP_MASK);
    
    return virt_base_addr;
}


void unmap_address(void * virt_base_addr, off_t base_addr) {

    void     *map_base; 

    // Undo virtual mapping from map_address()
    map_base = virt_base_addr - (base_addr & MAP_MASK);
    
    if (munmap(map_base,     MAP_SIZE) == -1) FATAL;
}



unsigned long read_write_mem(int fd, int argc, unsigned int address, char type, unsigned long writeval) {
    void          *virt_addr; 
    unsigned long  read_result = 0;
    int            access_type = 'w';
    
    if (argc < 1) { 
        exit(1); 
    } else {
        access_type = tolower(type);        
    }

    virt_addr = map_address(fd, (off_t) address);
        
    if (argc <= 2) {
        switch (access_type) {
            case 'b':
                read_result = *((unsigned char *) virt_addr);
                break;
            case 'h':
                read_result = *((unsigned short *) virt_addr);
                break;
            case 'w':
                read_result = *((unsigned long *) virt_addr);
                break;
            default:
                fprintf(stderr, "Illegal data type '%c'.\n", access_type);
                exit(2);
        }
        if (DEBUG) { printf("Value at address 0x%X (%p): 0x%X\n", address, virt_addr, read_result); }
        fflush(stdout);

    } else {
        switch (access_type) {
            case 'b':
                *((unsigned char *) virt_addr) = writeval;
                read_result = *((unsigned char *) virt_addr);
                break;
            case 'h':
                *((unsigned short *) virt_addr) = writeval;
                read_result = *((unsigned short *) virt_addr);
                break;
            case 'w':
                *((unsigned long *) virt_addr) = writeval;
                read_result = *((unsigned long *) virt_addr);
                break;
        }
        if (DEBUG) { printf("Written 0x%X at 0x%X; readback 0x%X\n", writeval, virt_addr, read_result); }
        fflush(stdout);
    }
    
    unmap_address(virt_addr, (off_t) address);    
    
    return read_result;
}



// ----------------------------------------------------------------------------
// PHY Functions
// ----------------------------------------------------------------------------

void print_phy_config(unsigned int base_addr, unsigned int phy_addr) {
    unsigned int   status;
    unsigned short phy_value;
    
    // All Registers probed are standard IEEE registers and should 
    // be supported by all PHYs.  Please refer to the documentation
    // for your specific PHY for additional debug capabilities or 
    // additional registers to be probed.

    status = MDIOPhyRegRead(base_addr, phy_addr, 0x00, &phy_value);
    printf("\n\rPHY Ctrl           : 0x%04x", phy_value);

    status = MDIOPhyRegRead(base_addr, phy_addr, 0x01, &phy_value);
    printf("\n\rPHY Status         : 0x%04x", phy_value);

    status = MDIOPhyRegRead(base_addr, phy_addr, 0x02, &phy_value);
    printf("\n\rPHY ID 1           : 0x%04x", phy_value);
    
    status = MDIOPhyRegRead(base_addr, phy_addr, 0x03, &phy_value);
    printf("\n\rPHY ID 2           : 0x%04x", phy_value);
    
    status = MDIOPhyRegRead(base_addr, phy_addr, 0x04, &phy_value);
    printf("\n\rPHY Auto Neg 0     : 0x%04x", phy_value);

    status = MDIOPhyRegRead(base_addr, phy_addr, 0x05, &phy_value);
    printf("\n\rPHY Link Ability  0: 0x%04x", phy_value);

    status = MDIOPhyRegRead(base_addr, phy_addr, 0x06, &phy_value);
    printf("\n\rPHY Auto Neg 1     : 0x%04x", phy_value);

    status = MDIOPhyRegRead(base_addr, phy_addr, 0x09, &phy_value);
    printf("\n\rPHY Gig Ctrl       : 0x%04x", phy_value);

    status = MDIOPhyRegRead(base_addr, phy_addr, 0x0A, &phy_value);
    printf("\n\rPHY Gig Status     : 0x%04x", phy_value);

    status = MDIOPhyRegRead(base_addr, phy_addr, 0x0F, &phy_value);
    printf("\n\rPHY Ext Status     : 0x%04x", phy_value);

    printf("\n");
}


void set_phy_control(unsigned int base_addr, unsigned int phy_addr, unsigned int loopback, unsigned int speed) {
    unsigned int   status;
    unsigned short phy_value = 0x0;

    // Uses IEEE standard register to put PHY in digital loopback mode.
    // Will set the appropriate bit to set the PHY to the requested speed.


    if (loopback) {
        phy_value |= ETH_CONTROL_LOOPBACK;
        
        switch (speed) {
            case   10:  phy_value |= ETH_CONTROL_SPEED_10;   break;
            case  100:  phy_value |= ETH_CONTROL_SPEED_100;  break;
            case 1000:  phy_value |= ETH_CONTROL_SPEED_1000; break;
        }
    } else {
        phy_value |= ETH_CONTROL_DEFAULT;
    }

    MDIOPhyRegWrite(base_addr, phy_addr, 0x00, phy_value);

    // Read back PHY values
    print_phy_config(base_addr, phy_addr);
}

